@@ -20,10 +20,13 @@
 
 #include <gtk/gtk.h>
 #include <glib/gi18n.h>
+#include <gdk/gdkkeysyms.h>
 #include <panel-applet.h>
 #include "config.h"
 #include "grdcpixmaps.h"
 #include "grdcappletmenuitem.h"
+#include "grdcstringarray.h"
+#include "grdcavahi.h"
 
 #define GRDC_APPLET_FACTORY_IID "OAFIID:Grdc_Applet_Factory"
 #define GRDC_APPLET_IID         "OAFIID:Grdc_Applet"
@@ -35,17 +38,31 @@ typedef struct _GrdcAppletData
     PanelApplet *applet;
 
     GtkWidget *image;
+    GtkWidget *popup_menu;
+    guint32 popup_time;
+
+    gint prev_size;
+
+    gchar *menu_group;
+    gint menu_group_count;
+    GtkWidget *menu_group_widget;
+
+    GrdcAvahi *avahi;
 } GrdcAppletData;
 
 typedef enum
 {
     GRDC_LAUNCH_MAIN,
+    GRDC_LAUNCH_PREF,
     GRDC_LAUNCH_QUICK,
-    GRDC_LAUNCH_FILE
+    GRDC_LAUNCH_FILE,
+    GRDC_LAUNCH_EDIT,
+    GRDC_LAUNCH_NEW
 } GrdcLaunchType;
 
 static void
-grdc_applet_launcher (GrdcLaunchType launch_type, const gchar *filename)
+grdc_applet_launcher (GrdcLaunchType launch_type, const gchar *filename,
+    const gchar *server, const gchar *protocol)
 {
     gint argc;
     gchar *argv[50];
@@ -62,6 +79,11 @@ grdc_applet_launcher (GrdcLaunchType launch_type, const gchar *filename)
     case GRDC_LAUNCH_MAIN:
         break;
 
+    case GRDC_LAUNCH_PREF:
+        argv[argc++] = g_strdup ("-p");
+        argv[argc++] = g_strdup ("3");
+        break;
+
     case GRDC_LAUNCH_QUICK:
         argv[argc++] = g_strdup ("-q");
         break;
@@ -70,7 +92,28 @@ grdc_applet_launcher (GrdcLaunchType launch_type, const gchar *filename)
         argv[argc++] = g_strdup ("-c");
         argv[argc++] = g_strdup (filename);
         break;
-    }    
+
+    case GRDC_LAUNCH_EDIT:
+        argv[argc++] = g_strdup ("-e");
+        argv[argc++] = g_strdup (filename);
+        break;
+
+    case GRDC_LAUNCH_NEW:
+        argv[argc++] = g_strdup ("-n");
+        break;
+
+    }
+
+    if (server)
+    {
+        argv[argc++] = g_strdup ("-s");
+        argv[argc++] = g_strdup (server);
+    }
+    if (protocol)
+    {
+        argv[argc++] = g_strdup ("-t");
+        argv[argc++] = g_strdup (protocol);
+    }
 
     argv[argc++] = NULL;
 
@@ -90,141 +133,14 @@ grdc_applet_launcher (GrdcLaunchType launch_type, const gchar *filename)
     {
         dialog = gtk_message_dialog_new (NULL,
             GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE,
-            error->message);
+            error->message, NULL);
         gtk_dialog_run (GTK_DIALOG (dialog));
         gtk_widget_destroy (dialog);
     }
 }
 
-static void
-grdc_applet_destroy (GtkWidget *widget, GrdcAppletData *appdata)
-{
-    g_free (appdata);
-}
-
-static void
-grdc_applet_change_size (GtkWidget *w, gint size, GrdcAppletData *appdata)
-{
-	GdkPixbuf *scaled;
-
-	scaled = gdk_pixbuf_scale_simple (grdc_pixmap_logo (), size - 2, size - 2, GDK_INTERP_HYPER);
-	gtk_image_set_from_pixbuf (GTK_IMAGE (appdata->image), scaled);
-	g_object_unref (scaled);
-	gtk_widget_set_size_request (GTK_WIDGET (appdata->image), MAX (11, size), MAX (11, size));
-}
-
-static void
-grdc_applet_menu_open_main (GrdcAppletMenuItem *menuitem, gpointer data)
-{
-    grdc_applet_launcher (GRDC_LAUNCH_MAIN, NULL);
-}
-
-static void
-grdc_applet_menu_open_quick (GrdcAppletMenuItem *menuitem, gpointer data)
-{
-    grdc_applet_launcher (GRDC_LAUNCH_QUICK, NULL);
-}
-
-static void
-grdc_applet_menu_open_file (GrdcAppletMenuItem *menuitem, gpointer data)
-{
-    grdc_applet_launcher (GRDC_LAUNCH_FILE, menuitem->filename);
-}
-
-static void
-grdc_applet_popup_menu_position (GtkMenu *menu, gint *x, gint *y, gboolean *push_in, gpointer data)
-{
-    GtkRequisition req;
-    gint tx, ty, size;
-    GrdcAppletData *appdata;
-
-    appdata = (GrdcAppletData*) data;
-    size = panel_applet_get_size (appdata->applet);
-    gdk_window_get_origin (GTK_WIDGET (appdata->applet)->window, &tx, &ty);
-    gtk_widget_size_request (GTK_WIDGET (menu), &req);
-
-    switch (panel_applet_get_orient (appdata->applet))
-    {
-    case PANEL_APPLET_ORIENT_UP:
-        ty -= req.height;
-        break;
-    case PANEL_APPLET_ORIENT_DOWN:
-        ty += size;
-        break;
-    case PANEL_APPLET_ORIENT_LEFT:
-        tx -= req.width;
-        break;
-    case PANEL_APPLET_ORIENT_RIGHT:
-        tx += size;
-        break;
-    }
-
-    *x = tx;
-    *y = ty;
-    *push_in = TRUE;
-}
-
-static void
-grdc_applet_popdown_menu (GtkWidget *widget, GrdcAppletData *appdata)
-{
-    gtk_widget_set_state (GTK_WIDGET (appdata->applet), GTK_STATE_NORMAL);
-}
-
-static gchar *grdc_applet_menu_group = NULL;
-
-static void
-grdc_applet_popup_menu_add_item (gpointer data, gpointer user_data)
-{
-    GtkWidget *group_item;
-    GtkWidget *hbox;
-    GtkWidget *widget;
-    GrdcAppletMenuItem *item;
-
-    item = GRDC_APPLET_MENU_ITEM (data);
-
-    if (item->group && item->group[0] != '\0')
-    {
-        if (!grdc_applet_menu_group || g_strcmp0 (grdc_applet_menu_group, item->group) != 0)
-        {
-            grdc_applet_menu_group = item->group;
-
-            group_item = gtk_menu_item_new ();
-            gtk_widget_show (group_item);
-            gtk_widget_set_sensitive (group_item, FALSE);
-            gtk_menu_append (GTK_MENU (user_data), group_item);
-
-            hbox = gtk_hbox_new (FALSE, 2);
-            gtk_widget_show (hbox);
-            gtk_container_add (GTK_CONTAINER (group_item), hbox);
-
-            widget = gtk_hseparator_new ();
-            gtk_widget_show (widget);
-            gtk_box_pack_start (GTK_BOX (hbox), widget, TRUE, TRUE, 0);
-
-            widget = gtk_label_new (grdc_applet_menu_group);
-            gtk_widget_show (widget);
-            gtk_box_pack_start (GTK_BOX (hbox), widget, FALSE, FALSE, 0);
-
-            widget = gtk_hseparator_new ();
-            gtk_widget_show (widget);
-            gtk_box_pack_start (GTK_BOX (hbox), widget, TRUE, TRUE, 0);
-        }
-    }
-    else
-    {
-        if (grdc_applet_menu_group != NULL)
-        {
-            grdc_applet_menu_group = NULL;
-            group_item = gtk_separator_menu_item_new ();
-            gtk_menu_append (GTK_MENU (user_data), group_item);
-            gtk_widget_show (group_item);
-        }
-    }
-    gtk_menu_append (GTK_MENU (user_data), GTK_WIDGET (data));
-}
-
 static gchar*
-grdc_applet_get_pref (const gchar *key)
+grdc_applet_get_pref_string (const gchar *key)
 {
     gchar filename[MAX_PATH_LEN];
     GKeyFile *gkeyfile;
@@ -245,10 +161,229 @@ grdc_applet_get_pref (const gchar *key)
     return value;
 }
 
+static gboolean
+grdc_applet_get_pref_boolean (const gchar *key)
+{
+    gchar filename[MAX_PATH_LEN];
+    GKeyFile *gkeyfile;
+    gboolean value;
+
+    g_snprintf (filename, sizeof (filename), "%s/.grdc/grdc.pref", g_get_home_dir ());
+    gkeyfile = g_key_file_new ();
+    if (g_key_file_load_from_file (gkeyfile, filename, G_KEY_FILE_NONE, NULL))
+    {
+        value = g_key_file_get_boolean (gkeyfile, "grdc_pref", key, NULL);
+    }
+    else
+    {
+        value = FALSE;
+    }
+    g_key_file_free (gkeyfile);
+
+    return value;
+}
+
+static void
+grdc_applet_set_pref_boolean (const gchar *key, gboolean value)
+{
+    gchar filename[MAX_PATH_LEN];
+    GKeyFile *gkeyfile;
+    gchar *content;
+    gsize length;
+
+    g_snprintf (filename, sizeof (filename), "%s/.grdc/grdc.pref", g_get_home_dir ());
+    gkeyfile = g_key_file_new ();
+    if (g_key_file_load_from_file (gkeyfile, filename, G_KEY_FILE_NONE, NULL))
+    {
+        g_key_file_set_boolean (gkeyfile, "grdc_pref", key, value);
+        content = g_key_file_to_data (gkeyfile, &length, NULL);
+        g_file_set_contents (filename, content, length, NULL);
+        g_free (content);
+    }
+    g_key_file_free (gkeyfile);
+}
+
+static void
+grdc_applet_destroy (GtkWidget *widget, GrdcAppletData *appdata)
+{
+    grdc_avahi_free (appdata->avahi);
+    g_free (appdata);
+}
+
+static void
+grdc_applet_size_allocate (GtkWidget *widget, GtkAllocation *allocation, GrdcAppletData *appdata)
+{
+    GdkPixbuf *scaled;
+    gint size = 0;
+
+    switch (panel_applet_get_orient (appdata->applet))
+    {
+    case PANEL_APPLET_ORIENT_UP:
+    case PANEL_APPLET_ORIENT_DOWN:
+        size = allocation->height;
+        break;
+    case PANEL_APPLET_ORIENT_LEFT:
+    case PANEL_APPLET_ORIENT_RIGHT:
+        size = allocation->width;
+        break;
+    }
+
+    if (size < 10) return;
+    if (size == appdata->prev_size) return;
+
+    appdata->prev_size = size;
+
+    scaled = grdc_pixmap (GRDC_PIXMAP_TYPE_LOGO, size);
+    gtk_image_set_from_pixbuf (GTK_IMAGE (appdata->image), scaled);
+    g_object_unref (scaled);
+}
+
+static void
+grdc_applet_menu_open_main (GrdcAppletMenuItem *menuitem, gpointer data)
+{
+    grdc_applet_launcher (GRDC_LAUNCH_MAIN, NULL, NULL, NULL);
+}
+
+static void
+grdc_applet_menu_open_pref (GrdcAppletMenuItem *menuitem, gpointer data)
+{
+    grdc_applet_launcher (GRDC_LAUNCH_PREF, NULL, NULL, NULL);
+}
+
+static void
+grdc_applet_menu_open_quick (GrdcAppletMenuItem *menuitem, GdkEventButton *event, GrdcAppletData *appdata)
+{
+    grdc_applet_launcher (GRDC_LAUNCH_QUICK, NULL, NULL, NULL);
+}
+
+static void
+grdc_applet_menu_open_file (GrdcAppletMenuItem *menuitem, GdkEventButton *event, GrdcAppletData *appdata)
+{
+    grdc_applet_launcher (event->button == 3 ? GRDC_LAUNCH_EDIT : GRDC_LAUNCH_FILE, menuitem->filename, NULL, NULL);
+}
+
+static void
+grdc_applet_menu_open_discovered (GrdcAppletMenuItem *menuitem, GdkEventButton *event, GrdcAppletData *appdata)
+{
+    grdc_applet_launcher (event->button == 3 ? GRDC_LAUNCH_NEW : GRDC_LAUNCH_QUICK, NULL, menuitem->name, menuitem->protocol);
+}
+
+static void
+grdc_applet_popdown_menu (GtkWidget *widget, GrdcAppletData *appdata)
+{
+    appdata->popup_menu = NULL;
+    gtk_widget_set_state (GTK_WIDGET (appdata->applet), GTK_STATE_NORMAL);
+    if (gtk_get_current_event_time () - appdata->popup_time <= 500)
+    {
+        grdc_applet_menu_open_main (NULL, NULL);
+    }
+}
+
+static void
+grdc_applet_popup_menu_position (GtkMenu *menu, gint *x, gint *y, gboolean *push_in, gpointer data)
+{
+    GdkScreen *screen;
+    GtkRequisition req;
+    gint tx, ty, size;
+    GrdcAppletData *appdata;
+
+    appdata = (GrdcAppletData*) data;
+    size = GTK_WIDGET (appdata->image)->allocation.width;
+    gdk_window_get_origin (GTK_WIDGET (appdata->applet)->window, &tx, &ty);
+    gtk_widget_size_request (GTK_WIDGET (menu), &req);
+
+    switch (panel_applet_get_orient (appdata->applet))
+    {
+    case PANEL_APPLET_ORIENT_UP:
+        ty -= req.height;
+        break;
+    case PANEL_APPLET_ORIENT_DOWN:
+        ty += size;
+        break;
+    case PANEL_APPLET_ORIENT_LEFT:
+        tx -= req.width;
+        break;
+    case PANEL_APPLET_ORIENT_RIGHT:
+        tx += size;
+        break;
+    }
+
+    screen = gdk_screen_get_default ();
+    tx = MIN (gdk_screen_get_width (screen) - req.width - 1, tx);
+    ty = MIN (gdk_screen_get_height (screen) - req.height - 1, ty);
+
+    *x = tx;
+    *y = ty;
+    *push_in = TRUE;
+}
+
+static void
+grdc_applet_popup_menu_update_group (GrdcAppletData* appdata)
+{
+    gchar *label;
+
+    if (appdata->menu_group)
+    {
+        if (!grdc_applet_get_pref_boolean ("applet_hide_count"))
+        {
+            label = g_strdup_printf ("%s (%i)", appdata->menu_group, appdata->menu_group_count);
+            gtk_label_set_text (GTK_LABEL (gtk_bin_get_child (GTK_BIN (appdata->menu_group_widget))), label);
+            g_free (label);
+        }
+        appdata->menu_group = NULL;
+        appdata->menu_group_count = 0;
+    }
+}
+
+static void
+grdc_applet_popup_menu_add_item (gpointer data, gpointer user_data)
+{
+    GrdcAppletData *appdata;
+    GrdcAppletMenuItem *item;
+    GtkWidget *submenu;
+    GtkWidget *image;
+
+    appdata = (GrdcAppletData*) user_data;
+    item = GRDC_APPLET_MENU_ITEM (data);
+
+    if (item->group && item->group[0] != '\0')
+    {
+        if (!appdata->menu_group || g_strcmp0 (appdata->menu_group, item->group) != 0)
+        {
+            grdc_applet_popup_menu_update_group (appdata);
+
+            appdata->menu_group = item->group;
+
+            appdata->menu_group_widget = gtk_image_menu_item_new_with_label (appdata->menu_group);
+            gtk_widget_show (appdata->menu_group_widget);
+            gtk_menu_shell_append (GTK_MENU_SHELL (appdata->popup_menu), appdata->menu_group_widget);
+
+            image = gtk_image_new_from_icon_name ((item->item_type == GRDC_APPLET_MENU_ITEM_DISCOVERED ?
+                "folder-remote" : "folder"), GTK_ICON_SIZE_MENU);
+            gtk_widget_show (image);
+            gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (appdata->menu_group_widget), image);
+
+            submenu = gtk_menu_new ();
+            gtk_widget_show (submenu);
+            gtk_menu_item_set_submenu (GTK_MENU_ITEM (appdata->menu_group_widget), submenu);
+        }
+        else
+        {
+            submenu = gtk_menu_item_get_submenu (GTK_MENU_ITEM (appdata->menu_group_widget));
+        }
+        gtk_menu_shell_append (GTK_MENU_SHELL (submenu), GTK_WIDGET (item));
+        appdata->menu_group_count++;
+    }
+    else
+    {
+        grdc_applet_popup_menu_update_group (appdata);
+        gtk_menu_shell_append (GTK_MENU_SHELL (appdata->popup_menu), GTK_WIDGET (item));
+    }
+}
+
 static void
 grdc_applet_popup_menu (GtkWidget *widget, GdkEventButton *event, GrdcAppletData *appdata)
 {
-    GtkWidget *menu;
     GtkWidget *menuitem;
     GPtrArray *menuitem_array;
     gint button, event_time;
@@ -257,44 +392,90 @@ grdc_applet_popup_menu (GtkWidget *widget, GdkEventButton *event, GrdcAppletData
     GDir *dir;
     gchar *groups;
     const gchar *name;
+    gboolean quick_ontop;
+    GHashTableIter iter;
+    gchar *tmp;
 
-    menu = gtk_menu_new ();
+    quick_ontop = grdc_applet_get_pref_boolean ("applet_quick_ontop");
+
+    appdata->popup_time = gtk_get_current_event_time ();
+    appdata->popup_menu = gtk_menu_new ();
+
+    if (quick_ontop)
+    {
+        menuitem = grdc_applet_menu_item_new (GRDC_APPLET_MENU_ITEM_QUICK);
+        g_signal_connect (menuitem, "button-press-event",
+            G_CALLBACK (grdc_applet_menu_open_quick), appdata);
+        gtk_widget_show (menuitem);
+        gtk_menu_shell_append (GTK_MENU_SHELL (appdata->popup_menu), menuitem);
+
+        menuitem = gtk_separator_menu_item_new ();
+        gtk_widget_show (menuitem);
+        gtk_menu_shell_append (GTK_MENU_SHELL (appdata->popup_menu), menuitem);
+    }
 
     g_snprintf (dirname, MAX_PATH_LEN, "%s/.grdc", g_get_home_dir ());
     dir = g_dir_open (dirname, 0, NULL);
     if (dir != NULL)
     {
         menuitem_array = g_ptr_array_new ();
+
+        /* Iterate all remote desktop profiles */
         while ((name = g_dir_read_name (dir)) != NULL)
         {
             if (!g_str_has_suffix (name, ".grdc")) continue;
             g_snprintf (filename, MAX_PATH_LEN, "%s/%s", dirname, name);
 
             menuitem = grdc_applet_menu_item_new (GRDC_APPLET_MENU_ITEM_FILE, filename);
-	        g_signal_connect (G_OBJECT (menuitem), "activate", G_CALLBACK (grdc_applet_menu_open_file), NULL);
-	        gtk_widget_show (menuitem);
+            g_signal_connect (G_OBJECT (menuitem), "button-press-event",
+                G_CALLBACK (grdc_applet_menu_open_file), appdata);
+            gtk_widget_show (menuitem);
 
             g_ptr_array_add (menuitem_array, menuitem);
         }
 
-        groups = grdc_applet_get_pref ("groups");
+        /* Iterate all discovered services from Avahi */
+        if (appdata->avahi)
+        {
+            g_hash_table_iter_init (&iter, appdata->avahi->discovered_services);
+            while (g_hash_table_iter_next (&iter, NULL, (gpointer*) &tmp))
+            {
+                menuitem = grdc_applet_menu_item_new (GRDC_APPLET_MENU_ITEM_DISCOVERED, tmp);
+                g_signal_connect (G_OBJECT (menuitem), "button-press-event",
+                    G_CALLBACK (grdc_applet_menu_open_discovered), appdata);
+                gtk_widget_show (menuitem);
+
+                g_ptr_array_add (menuitem_array, menuitem);
+            }
+        }
+
+        tmp = grdc_applet_get_pref_string ("groups");
+        groups = g_strdup_printf ("%s%s%s", tmp, (tmp && tmp[0] ? "," : ""), _("Discovered"));
+        g_free (tmp);
         g_ptr_array_sort_with_data (menuitem_array, grdc_applet_menu_item_compare, groups);
         g_free (groups);
 
-        grdc_applet_menu_group = NULL;
-        g_ptr_array_foreach (menuitem_array, grdc_applet_popup_menu_add_item, menu);
+        appdata->menu_group = NULL;
+        appdata->menu_group_count = 0;
+        g_ptr_array_foreach (menuitem_array, grdc_applet_popup_menu_add_item, appdata);
+
+        grdc_applet_popup_menu_update_group (appdata);
 
         g_ptr_array_free (menuitem_array, FALSE);
     }
 
-    menuitem = gtk_separator_menu_item_new ();
-    gtk_menu_append (GTK_MENU (menu), menuitem);
-    gtk_widget_show (menuitem);
+    if (!quick_ontop)
+    {
+        menuitem = gtk_separator_menu_item_new ();
+        gtk_widget_show (menuitem);
+        gtk_menu_shell_append (GTK_MENU_SHELL (appdata->popup_menu), menuitem);
 
-    menuitem = grdc_applet_menu_item_new (GRDC_APPLET_MENU_ITEM_QUICK, NULL);
-    g_signal_connect (menuitem, "activate", G_CALLBACK (grdc_applet_menu_open_quick), NULL);
-    gtk_menu_append (GTK_MENU (menu), menuitem);
-    gtk_widget_show (menuitem);
+        menuitem = grdc_applet_menu_item_new (GRDC_APPLET_MENU_ITEM_QUICK);
+        g_signal_connect (menuitem, "button-press-event",
+            G_CALLBACK (grdc_applet_menu_open_quick), appdata);
+        gtk_widget_show (menuitem);
+        gtk_menu_shell_append (GTK_MENU_SHELL (appdata->popup_menu), menuitem);
+    }
 
     if (event)
     {
@@ -308,11 +489,12 @@ grdc_applet_popup_menu (GtkWidget *widget, GdkEventButton *event, GrdcAppletData
     }
 
     gtk_widget_set_state (GTK_WIDGET (appdata->applet), GTK_STATE_SELECTED);
-    g_signal_connect (menu, "deactivate", G_CALLBACK (grdc_applet_popdown_menu), appdata);
+    g_signal_connect (G_OBJECT (appdata->popup_menu), "deactivate", G_CALLBACK (grdc_applet_popdown_menu), appdata);
 
-    gtk_menu_attach_to_widget (GTK_MENU (menu), widget, NULL);
-    gtk_widget_realize (menu);
-    gtk_menu_popup (GTK_MENU (menu), NULL, NULL, grdc_applet_popup_menu_position, appdata, button, event_time);
+    gtk_menu_attach_to_widget (GTK_MENU (appdata->popup_menu), widget, NULL);
+    gtk_widget_realize (appdata->popup_menu);
+    gtk_menu_popup (GTK_MENU (appdata->popup_menu), NULL, NULL,
+        grdc_applet_popup_menu_position, appdata, button, event_time);
 }
 
 static gboolean
@@ -323,10 +505,7 @@ grdc_applet_button_press_event (GtkWidget *widget, GdkEventButton *event, GrdcAp
         switch (event->type)
         {
         case GDK_BUTTON_PRESS:
-            grdc_applet_popup_menu (widget,  event, appdata);
-            break;
-        case GDK_2BUTTON_PRESS:
-            grdc_applet_menu_open_main (NULL, NULL);
+            grdc_applet_popup_menu (widget, event, appdata);
             break;
         default:
             break;
@@ -336,6 +515,29 @@ grdc_applet_button_press_event (GtkWidget *widget, GdkEventButton *event, GrdcAp
     return FALSE;
 }
 
+static void
+grdc_applet_menu_enable_avahi (BonoboUIComponent *uic,
+    const char                  *path,
+    Bonobo_UIComponent_EventType type,
+    const char                  *state,
+    gpointer                     userdata)
+{
+    GrdcAppletData *appdata = (GrdcAppletData*) userdata;
+
+    if (!appdata->avahi) return;
+
+    if (state[0] == '1')
+    {
+        grdc_applet_set_pref_boolean ("applet_enable_avahi", TRUE);
+        if (!appdata->avahi->started) grdc_avahi_start (appdata->avahi);
+    }
+    else
+    {
+        grdc_applet_set_pref_boolean ("applet_enable_avahi", FALSE);
+        grdc_avahi_stop (appdata->avahi);
+    }
+}
+
 static void
 grdc_applet_menu_about (BonoboUIComponent *uic, GrdcAppletData *appdata, const char *verb)
 {
@@ -370,7 +572,7 @@ grdc_applet_menu_about (BonoboUIComponent *uic, GrdcAppletData *appdata, const c
         "copyright", "Copyright (C) 2009 - Vic Lee",
         "license", license,
         "wrap-license", TRUE,
-        "logo", grdc_pixmap_logo (),
+        "logo", GRDC_PIXMAP_LOGO,
         "website", "http://grdc.sourceforge.net",
         NULL);
 
@@ -379,26 +581,37 @@ grdc_applet_menu_about (BonoboUIComponent *uic, GrdcAppletData *appdata, const c
 
 static const BonoboUIVerb grdc_applet_menu_verbs [] =
 {   
-	BONOBO_UI_UNSAFE_VERB ("OpenMain", grdc_applet_menu_open_main),
-	BONOBO_UI_UNSAFE_VERB ("About", grdc_applet_menu_about),
-	BONOBO_UI_VERB_END
+    BONOBO_UI_UNSAFE_VERB ("OpenMain", grdc_applet_menu_open_main),
+    BONOBO_UI_UNSAFE_VERB ("OpenPref", grdc_applet_menu_open_pref),
+    BONOBO_UI_UNSAFE_VERB ("About", grdc_applet_menu_about),
+    BONOBO_UI_VERB_END
 };
 
 static const gchar *grdc_applet_menu_xml = 
 "    <popup name='button3'>"
 "      <menuitem name='OpenMain' verb='OpenMain' _label='%s' pixtype='stock' pixname='gtk-execute'  />"
+"      <menuitem name='OpenPref' verb='OpenPref' _label='%s' pixtype='stock' pixname='gtk-preferences'  />"
 "      <menuitem name='About'    verb='About'    _label='%s' pixtype='stock' pixname='gtk-about' />"
+#ifdef HAVE_LIBAVAHI_CLIENT
+"      <separator />"
+"      <menuitem name='EnableAvahi' type='toggle' id='EnableAvahi' _label='%s' />"
+#endif
 "    </popup>"
 ;
 
 static void
 grdc_applet_create (PanelApplet *applet)
 {
+    BonoboUIComponent *popup_component;
     gchar buf[1000];
     GrdcAppletData *appdata;
 
     appdata = g_new0 (GrdcAppletData, 1);
     appdata->applet = applet;
+    appdata->popup_menu = NULL;
+    appdata->popup_time = 0;
+    appdata->prev_size = 0;
+    appdata->avahi = grdc_avahi_new ();
 
     appdata->image = gtk_image_new_from_pixbuf (NULL);
     gtk_widget_show (appdata->image);
@@ -406,15 +619,30 @@ grdc_applet_create (PanelApplet *applet)
     gtk_container_add (GTK_CONTAINER (applet), appdata->image);
 
     g_signal_connect (GTK_WIDGET (applet), "destroy", G_CALLBACK (grdc_applet_destroy), appdata);
-    g_signal_connect (GTK_WIDGET (applet), "change-size", G_CALLBACK (grdc_applet_change_size), appdata);
     g_signal_connect (GTK_WIDGET (applet), "button-press-event", G_CALLBACK (grdc_applet_button_press_event), appdata);
+    g_signal_connect (appdata->image, "size-allocate", G_CALLBACK (grdc_applet_size_allocate), appdata);
 
-    g_snprintf (buf, sizeof (buf), grdc_applet_menu_xml, _("Open Main Window"), _("About"));
+    g_snprintf (buf, sizeof (buf), grdc_applet_menu_xml, _("Open Main Window"), _("Preferences"),
+        _("About")
+#ifdef HAVE_LIBAVAHI_CLIENT
+        , _("Enable Service Discovery")
+#endif
+        );
     panel_applet_setup_menu (applet, buf, grdc_applet_menu_verbs, appdata);
 
-    grdc_applet_change_size (GTK_WIDGET (applet), panel_applet_get_size (applet), appdata);
+    popup_component = panel_applet_get_popup_component (applet);
+    bonobo_ui_component_add_listener (popup_component, "EnableAvahi", grdc_applet_menu_enable_avahi, appdata);
+
+    panel_applet_set_flags (applet, PANEL_APPLET_EXPAND_MINOR);
+    panel_applet_set_background_widget (applet, GTK_WIDGET (applet));
 
     gtk_widget_show (GTK_WIDGET (applet));
+
+    if (grdc_applet_get_pref_boolean ("applet_enable_avahi"))
+    {
+        bonobo_ui_component_set_prop (popup_component, "/commands/EnableAvahi", "state", "1", NULL);
+        grdc_avahi_start (appdata->avahi);
+    }
 }
 
 static gboolean
@@ -426,6 +654,8 @@ grdc_applet_factory (PanelApplet *applet, const gchar *iid, gpointer data)
     bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
     textdomain (GETTEXT_PACKAGE);
 
+    grdc_pixmap_init ();
+
     grdc_applet_create (applet);
 
     return TRUE;
